/*-
 * Copyright (c) 2006 Verdens Gang AS
 * Copyright (c) 2006-2019 Varnish Software AS
 * All rights reserved.
 *
 * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Stuff that should *never* be exposed to a VMOD
 */

#include "cache.h"

#include "vsb.h"
#include "vapi/vst_int.h"

#include <sys/socket.h>

#include <string.h>
#include <limits.h>
#include <unistd.h>

#include "common/common_param.h"

#ifdef NOT_IN_A_VMOD
#  include "VSC_main.h"
#endif

/* -------------------------------------------------------------------*/

struct vfp;
struct cli_proto;
struct msebytesorpct;
struct poolparam;

/*--------------------------------------------------------------------
 * HTTP Protocol connection structure
 *
 * This is the protocol independent object for a HTTP connection, used
 * both for backend and client sides.
 *
 */

struct http_conn {
	unsigned		magic;
#define HTTP_CONN_MAGIC		0x3e19edd1

	int			*rfd;
	enum sess_close		doclose;
	enum body_status	body_status;
	struct ws		*ws;
	char			*rxbuf_b;
	char			*rxbuf_e;
	char			*pipeline_b;
	char			*pipeline_e;
	ssize_t			content_length;
	void			*priv;
	const struct vco	*oper;
	void			*oper_priv;

	/* Timeouts */
	vtim_real		*t_first;
	vtim_dur		first_byte_timeout;
	vtim_dur		between_bytes_timeout;
	vtim_dur		last_byte_timeout;
};

typedef enum htc_status_e htc_complete_f(struct http_conn *);

/* -------------------------------------------------------------------*/

extern volatile struct params * cache_param;

/* Prototypes etc ----------------------------------------------------*/

/* cache_acceptor.c */
void VCA_Init(void);
void VCA_Start(void);
void VCA_Shutdown(void);
unsigned SES_Listening(const struct sess *);

/* cache_backend_cfg.c */
void VBE_InitCfg(void);
void VBE_Poll(void);

/* cache_ban.c */

/* for stevedoes resurrecting bans */
typedef void ban_iterate_f(const uint8_t *ban, unsigned len, void *priv);
void BAN_Iterate(ban_iterate_f *func, void *priv);
void BAN_Hold(void);
void BAN_Release(void);
void BAN_Reload(const uint8_t *ban, unsigned len);
struct ban *BAN_FindBan(double t0);
struct ban *BAN_FirstBan(void);
void BAN_RefBan(struct objcore *oc, struct ban *);
int BAN_RefBanBatch(struct objcore_vtqhead *head, int try);
int BAN_RefBanMulti(struct objcore_vtqhead *head, struct ban *, unsigned,
    int try);
double BAN_Time(const struct ban *ban);
unsigned BAN_SpecLen(const uint8_t *banspec);

/* cache_busyobj.c */
struct busyobj *VBO_GetBusyObj(struct worker *, const struct req *);
void VBO_ReleaseBusyObj(struct worker *wrk, struct busyobj **busyobj);

/* cache_director.c */
int VDI_GetHdr(struct worker *, struct busyobj *);
int VDI_GetBody(struct worker *, struct busyobj *);
const struct suckaddr *VDI_GetIP(struct worker *, struct busyobj *);
void VDI_Finish(struct worker *wrk, struct busyobj *bo);
enum sess_close VDI_Http1Pipe(struct req *, struct busyobj *);
void VDI_Panic(const struct director *, struct vsb *, const char *nm);
void VDI_Event(const struct director *d, enum vcl_event_e ev);
void VDI_Init(void);

/* cache_exp.c */
double EXP_Ttl(const struct req *, const struct objcore *);
double EXP_Ttl_grace(const struct req *, const struct objcore *oc);
void EXP_RefNewObjcore(struct objcore *);
void EXP_Insert(struct worker *wrk, struct objcore *oc);
void EXP_ImportBatchAdd(struct objcore_vstqhead *head, struct objcore *oc);
int EXP_ImportBatch(struct worker *wrk, struct objcore_vstqhead *head,
    unsigned n, int try);
void EXP_Remove(struct objcore *, const struct objcore *);

#define EXP_Dttl(req, oc) (oc->ttl - (req->t_req - oc->t_origin))

/* cache_backend_ssl.c */
void BSSL_Init(void);

/* cache_tls.c */
void VTLS_Init(void);
void VTLS_NewPool(struct pool *pp);
void VTLS_del_sess(struct pool *pp, void **tsp);
void VTLS_vsl_set(void *p, struct vsl_log *vsl);
const struct vco *VTLS_conn_oper_client(void *priv, void **ppriv);
const struct vco *VTLS_conn_oper_backend(void *priv, void **ppriv);
void VTLS_Poll(void);

/* cache_expire.c */

/*
 * The set of variables which control object expiry are inconveniently
 * 24 bytes long (double+3*float) and this causes alignment waste if
 * we put then in a struct.
 * These three macros operate on the struct we don't use.
 */

#define EXP_ZERO(xx)							\
	do {								\
		(xx)->t_origin = 0.0;					\
		(xx)->ttl = 0.0;					\
		(xx)->grace = 0.0;					\
		(xx)->keep = 0.0;					\
	} while (0)

#define EXP_COPY(to,fm)							\
	do {								\
		(to)->t_origin = (fm)->t_origin;			\
		(to)->ttl = (fm)->ttl;					\
		(to)->grace = (fm)->grace;				\
		(to)->keep = (fm)->keep;				\
	} while (0)

#define EXP_WHEN(to)							\
	((to)->t_origin + (to)->ttl + (to)->grace + (to)->keep)

/* cache_exp.c */
unsigned EXP_Rearm(struct objcore *, double now, double ttl, double grace,
    double keep);
unsigned EXP_Reduce(struct objcore *oc, vtim_real now,
    vtim_dur ttl, vtim_dur grace, vtim_dur keep);

/* From cache_main.c */
void BAN_Init(void);
void BAN_Compile(void);
void BAN_Shutdown(void);

/* From cache_hash.c */
void BAN_NewObjCore(struct objcore *oc);
void BAN_DestroyObj(struct objcore *oc);
struct ban *BAN_TestObject(struct objcore *oc, uint64_t *pseq);
int BAN_CheckObject(struct worker *, struct objcore *, struct req *,
    struct ban **pbanref, uint64_t banseq);

/* cache_busyobj.c */
void VBO_NewPool(struct pool *);

/* cache_cli.c [CLI] */
void CLI_Init(void);
void CLI_Run(void);
void CLI_AddFuncs(struct cli_proto *p);

/* cache_deliver_proc.c */
void VDP_close(struct req *req);
int VDP_DeliverObj(struct req *req);

extern const struct vdp VDP_gzip;
extern const struct vdp VDP_gunzip;
extern const struct vdp VDP_esi;

/* cache_expire.c */
void EXP_Init(void);

/* cache_fetch.c */
enum vbf_fetch_mode_e {
	VBF_NORMAL = 0,
	VBF_PASS = 1,
	VBF_BACKGROUND = 2,
};
void VBF_Fetch(struct worker *wrk, struct req *req,
    struct objcore *oc, struct objcore *oldoc, enum vbf_fetch_mode_e);
void VBF_WaitStream(struct worker *wrk, struct req *req);

/* cache_fetch_proc.c */
void VFP_Init(void);
enum vfp_status VFP_GetStorage(struct vfp_ctx *, ssize_t *sz, uint8_t **ptr);
void VFP_Extend(const struct vfp_ctx *, ssize_t sz);
struct vfp_entry *VFP_Push(struct vfp_ctx *, const struct vfp *);
void VFP_Setup(struct vfp_ctx *vc, struct worker *wrk);
int VFP_Open(struct vfp_ctx *bo);
void VFP_Close(struct vfp_ctx *bo);

extern const struct vfp VFP_gunzip;
extern const struct vfp VFP_gzip;
extern const struct vfp VFP_testgunzip;
extern const struct vfp VFP_esi;
extern const struct vfp VFP_esi_gzip;


/* cache_http.c */
void HTTP_Init(void);

/* cache_http1_proto.c */

htc_complete_f HTTP1_Complete;
uint16_t HTTP1_DissectRequest(struct http_conn *, struct http *);
uint16_t HTTP1_DissectResponse(struct http_conn *, struct http *resp,
    const struct http *req);
unsigned HTTP1_Write(const struct worker *w, const struct http *hp, const int*);

/* cache_main.c */
extern pthread_key_t panic_key;

void THR_SetName(const char *name);
const char* THR_GetName(void);
void THR_SetBusyobj(const struct busyobj *);
struct busyobj * THR_GetBusyobj(void);
void THR_SetRequest(const struct req *);
struct req * THR_GetRequest(void);
void THR_Init(void);

/* cache_lck.c */
void LCK_Init(void);

/* cache_mempool.c */
void MPL_AssertSane(const void *item);
struct mempool * MPL_New(const char *name, volatile struct poolparam *pp,
    volatile unsigned *cur_size, struct pool *p);
void MPL_Destroy(struct mempool **mpp);
void *MPL_Get(struct mempool *mpl, unsigned *size);
void MPL_Free(struct mempool *mpl, void *item);

/* cache_obj.c */
void ObjInit(void);
struct objcore * ObjNew(const struct worker *);
void ObjDestroy(const struct worker *, struct objcore **);
int ObjGetSpace(struct worker *, struct objcore *, ssize_t *sz, uint8_t **ptr);
void ObjExtend(struct worker *, struct objcore *, ssize_t l);
uint64_t ObjWaitExtend(const struct worker *, const struct objcore *,
    uint64_t l);
void ObjSetState(struct worker *, struct objcore *, enum boc_state_e next);
void ObjWaitState(const struct objcore *, enum boc_state_e want);
void ObjTrimStore(struct worker *, struct objcore *);
void ObjFinish(struct worker *, struct objcore *);
void ObjTouch(struct worker *, struct objcore *, vtim_real now);
void ObjFreeObj(struct worker *, struct objcore *);
void ObjSlim(struct worker *, struct objcore *);
void *ObjSetAttr(struct worker *, struct objcore *, oaref_t,
    enum obj_attr, ssize_t len, const void *);
int ObjCopyAttr(struct worker *, struct objcore *, oaref_t,
    struct objcore *, oaref_t, enum obj_attr attr);
oaauxref_t ObjSetAuxAttr(struct worker *, struct objcore *, oaref_t,
    enum obj_attr, ssize_t len, uint8_t **pdata);
void ObjBocDone(struct worker *, struct objcore *, struct boc **);
int ObjCheckVary(struct worker *, struct objcore *, struct http *http);
int ObjGetVary(struct worker *, struct objcore *, uint8_t *buf, unsigned buflen,
    uint64_t *pvaryval);
int ObjSetVary(struct worker *, struct objcore *, oaref_t, const uint8_t *vary,
    uint64_t varyval);
void ObjSetLen(struct worker *, struct objcore *, oaref_t, uint64_t);
int ObjSetSlicerData(struct worker *wrk, struct objcore *oc, oaref_t oaref,
    const uint8_t *data);
int ObjGetSlicerData(struct worker *wrk, struct objcore *oc, oaref_t opt_oaref,
    uint8_t *data);

int ObjSetDouble(struct worker *, struct objcore *, oaref_t,
    enum obj_attr, double);
int ObjSetU32(struct worker *, struct objcore *, oaref_t,
    enum obj_attr, uint32_t);
int ObjSetU64(struct worker *, struct objcore *, oaref_t,
    enum obj_attr, uint64_t);

void ObjSetFlag(struct worker *, struct objcore *, oaref_t,
    enum obj_flags of, int val);

void ObjSendEvent(struct worker *, struct objcore *oc, unsigned event);

#define OEV_INSERT	(1U<<1)
#define OEV_BANCHG	(1U<<2)
#define OEV_TTLCHG	(1U<<3)
#define OEV_EXPIRE	(1U<<4)

#define OEV_MASK (OEV_INSERT|OEV_BANCHG|OEV_TTLCHG|OEV_EXPIRE)

typedef void obj_event_f(struct worker *, void *priv, struct objcore *,
    unsigned);

uintptr_t ObjSubscribeEvents(obj_event_f *, void *, unsigned mask);
void ObjUnsubscribeEvents(uintptr_t *);
int ObjSubscribedEvents(unsigned mask);

unsigned ObjPickSlot(void);
void ObjLock(unsigned slot);
void ObjLockAssertHeld(unsigned slot);
int ObjTrylock(unsigned slot);
void ObjUnlock(unsigned slot);
int ObjCondWait(unsigned slot, double tmo);
void ObjCondBroadcast(unsigned slot);


/* cache_panic.c */
void PAN_Init(void);
int PAN_already(struct vsb *, const void *);
void pan_req(struct vsb *vsb, const struct req *req);
void pan_objcore(struct vsb *vsb, const char *typ, const struct objcore *oc);
const char *body_status_2str(enum body_status e);
const char *sess_close_2str(enum sess_close sc, int want_desc);

/* cache_pool.c */
void Pool_Init(void);
int Pool_Task(struct pool *pp, struct pool_task *task, enum task_prio prio,
    enum vst_step_e);
int Pool_Task_Arg(struct worker *, enum task_prio, task_func_t *,
    const void *arg, size_t arg_len);
void Pool_Sumstat(const struct worker *w);
int Pool_TrySumstat(const struct worker *wrk);
void Pool_PurgeStat(unsigned nobj);
int Pool_Task_Any(struct pool_task *task, enum task_prio prio, enum vst_step_e);

/* cache_range.c [VRG] */
void VRG_dorange(struct req *req, const char *r);
int VRG_gen_dorange(const char *r, ssize_t *range_low, ssize_t *range_high, ssize_t len, const char **err_code);

/* cache_req.c */
struct req *Req_New(const struct worker *, struct sess *);
void Req_Release(struct req *);
void Req_Rollback(struct req *req);
void Req_Cleanup(struct sess *sp, struct worker *wrk, struct req *req);
void Req_Fail(struct req *req, enum sess_close reason);
void Req_AcctLogCharge(struct VSC_main_wrk *, struct req *);
const char *Req_LogStart(const struct worker *, struct req *);
void Req_LogHit(struct worker *, struct req *, struct objcore *, intmax_t);

/* cache_req_body.c */
int VRB_Ignore(struct req *);
ssize_t VRB_Cache(struct req *, ssize_t maxsize);
ssize_t VRB_Iterate(struct req *, objiterate_f *func, void *priv);
void VRB_Free(struct req *);

/* cache_req_fsm.c [CNT] */

enum req_fsm_nxt {
	REQ_FSM_MORE,
	REQ_FSM_DONE,
	REQ_FSM_DISEMBARK,
};

void CNT_Embark(struct worker *, struct req *);
enum req_fsm_nxt CNT_Request(struct req *);

/* cache_session.c */
void SES_NewPool(struct pool *);
void SES_DestroyPool(struct pool *);
void SES_Wait(struct sess *, const struct transport *);
void SES_Ref(struct sess *sp);
void SES_Rel(struct sess *sp);

const char * HTC_Status(enum htc_status_e);
void HTC_RxInit(struct http_conn *htc, struct ws *ws);
void HTC_RxPipeline(struct http_conn *htc, void *);
enum htc_status_e HTC_RxStuff(struct http_conn *, htc_complete_f *,
    vtim_real *t1, vtim_real *t2, vtim_real ti, vtim_real tn, vtim_dur td,
    int maxbytes);
ssize_t HTC_Read(struct http_conn *htc, void *buf, size_t len);

#define SESS_ATTR(UP, low, typ, len)					\
	int SES_Set_##low(const struct sess *sp, const typ *src);	\
	int SES_Reserve_##low(struct sess *sp, typ **dst);
#include "tbl/sess_attr.h"
int SES_Set_String_Attr(struct sess *sp, enum sess_attr a, const char *src);


enum htc_status_e {
#define HTC_STATUS(e, n, s, l) HTC_S_ ## e = n,
#include "tbl/htc.h"
};

/* cache_shmlog.c */
extern struct VSC_main *VSC_C_main;
void VSM_Init(void);
void VSL_Setup(struct vsl_log *vsl, void *ptr, size_t len);
void VSL_ChgId(struct vsl_log *vsl, const char *typ, const char *why,
    uint32_t vxid);
void VSL_End(struct vsl_log *vsl);
void VSL_Flush(struct vsl_log *, int overflow);

/* cache_conn_pool.c */
struct conn_pool;
void VCP_Init(void);
void VCP_Panic(struct vsb *, struct conn_pool *);

/* cache_backend_probe.c */
void VBP_Init(void);

/* cache_vary.c */
unsigned VRY_Len(const uint8_t *vary);
uint64_t VRY_Compute(const uint8_t *vary, struct http *http);
int VRY_Create(struct busyobj *bo, uint8_t **pvary);

/* cache_vcl.c */
struct director *VCL_DefaultDirector(const struct vcl *);
const struct vrt_backend_probe *VCL_DefaultProbe(const struct vcl *);
void VCL_Init(void);
void VCL_Panic(struct vsb *, const struct vcl *);
void VCL_Poll(void);
void VCL_Ref(struct vcl *);
void VCL_Refresh(struct vcl **);
void VCL_Rel(struct vcl **);
void VCL_TaskEnter(struct vrt_privs *);
void VCL_TaskLeave(struct vrt_privs *);
struct vmod_priv *VCL_TaskPrivLookup(struct vrt_privs *, const struct vcl *,
    uintptr_t, const void *);
const char *VCL_Return_Name(unsigned);
const char *VCL_Method_Name(unsigned);
void VCL_Bo2Ctx(struct vrt_ctx *, struct busyobj *);
void VCL_Req2Ctx(struct vrt_ctx *, struct req *);

#define VCL_MET_MAC(l,u,t,b) \
    void VCL_##l##_method(struct vcl *, struct worker *, struct req *, \
	struct busyobj *bo, void *specific);
#include "tbl/vcl_returns.h"


typedef int vcl_be_func(struct cli *, struct director *, void *);

int VCL_IterDirector(struct cli *, const char *, vcl_be_func *, void *);

/* cache_vrt.c */
void pan_privs(struct vsb *, const struct vrt_privs *);

/* cache_vrt_priv.c */
extern struct vrt_privs cli_task_privs[1];

/* cache_vrt_re.c */
char *VPI_regsub2(VRT_CTX, int all, const char *in, void *re, const char *sub,
    int *perr);

/* cache_vrt_vmod.c */
void VMOD_Init(void);
void VMOD_Panic(struct vsb *);

#if defined(ENABLE_COVERAGE) || defined(ENABLE_SANITIZER)
#  define DONT_DLCLOSE_VMODS
#endif

/* cache_wrk.c */
void WRK_Init(void);
void WRK_AddStat(struct worker *);

/* cache_vst.c */
void VST_Init(struct pool *);
void VST_Fini(struct pool *);

void VST_Start(struct worker *);
void VST_Stop(struct worker *);
void VST_QueueEnter(struct pool *, struct pool_task *, enum vst_step_e);
void VST_QueueLeave(struct worker *, struct pool_task *);
void VST_Pause(struct worker *, struct pool_task *, enum vst_step_e);
void VST_Resume(struct worker *, struct pool_task *);
void VST_Schedule(struct worker *, struct pool_task *, enum vst_step_e,
    vtim_real);
void VST_Cancel(struct worker *, struct pool_task *);

void VST_Step(struct worker *, enum vst_step_e step);
void VST_TaskEnter(struct worker *, enum vst_step_e step, vtim_real ts);
void VST_TaskLeave(struct worker *);
void VST_TaskSwitch(struct worker *, struct pool_task **, enum vst_step_e,
    struct pool_task *);
void VST_Address(struct worker *wrk, const char *fmt, ...) v_printflike_(2, 3);
void VST_Description(struct worker *wrk, const char *fmt, ...) v_printflike_(2, 3);
void VST_TaskId(struct worker *, uint32_t xid, uint32_t pxid);

void VST_Request(struct worker *, const struct http *hp);
void VST_Suckaddr(struct worker *, const struct suckaddr *sa);

/* http1/cache_http1_pipe.c */
void V1P_Init(void);

/* cache_http2_deliver.c */
void V2D_Init(void);

/* cache/cache_governor.c */
void GOV_Init(void);
size_t GOV_GetTarget(void);
typedef int gov_nukeone_f(struct worker *, void *priv);
void GOV_Enable(gov_nukeone_f *func, void *priv, uint64_t *p_free_stat);
void *GOV_Alloc(struct worker *, size_t bytes, unsigned align);
void GOV_Free(void *);
void GOV_FlushCache(void);

/* stevedore.c */
void STV_open(void);
void STV_signal(int s, void *si);
void STV_start(void);
void STV_close(void);
const struct stevedore *STV_next(void);
int STV_BanInfoDrop(const uint8_t *ban, unsigned len);
int STV_BanInfoNew(const uint8_t *ban, unsigned len);
void STV_BanExport(const uint8_t *banlist, unsigned len);
struct stv_hints {
	unsigned		is_synth_client:1;
	unsigned		is_synth_backend:1;
	unsigned		is_req_body:1;
#define OBJ_VARATTR(U, l, n)			\
	unsigned		oause_##l:1;
#define OBJ_AUXATTR(U, l, n)			\
	unsigned		oause_##l:1;
#include "tbl/obj_attr.h"
#define OBJ_VARATTR(U, l, n)			\
	unsigned		oalen_##l;
#include "tbl/obj_attr.h"
	ssize_t			size_body;
	struct vrt_privs	*vrt_privs;
	uintptr_t		vrt_privs_id;
	struct vcl		*vcl;
};
enum stv_alloc_result STV_NewObject(struct worker *, struct objcore *,
    const struct stevedore *, const struct stv_hints *hints);

struct stv_buffer;
struct stv_buffer *STV_AllocBuf(struct worker *wrk, const struct stevedore *stv,
    size_t size);
void STV_FreeBuf(struct worker *wrk, struct stv_buffer **pstvbuf);
void *STV_GetBufPtr(struct stv_buffer *stvbuf, size_t *psize);

/* cache_accounting.c */
void ACCG_Init(void);
void ACCG_flush_req_stats(struct req *req);
void ACCG_drop_req_stats(struct req *req);
void ACCG_sub_ns_copy(struct req *sub, struct req *preq);
void ACCG_add_bo_stats(struct busyobj *bo);
void ACCG_flush_bo_stats(struct busyobj *bo);
void ACCG_drop_bo_stats(struct busyobj *bo);
void ACCG_bo_copy_conf(struct busyobj  *bo, struct req *req);

/* cache_smuggler.c */
void SMUG_Init(void);
int SMUG_Fence(uint64_t);

/* storage/mse_cache.c */
void MSE_Init(void);

#define FEATURE(x)	COM_FEATURE(cache_param->feature_bits, x)
#define EXPERIMENTAL(x)	COM_EXPERIMENTAL(cache_param->experimental_bits, x)
#define DO_DEBUG(x)	COM_DO_DEBUG(cache_param->debug_bits, x)

#define DSL(debug_bit, id, ...)					\
	do {							\
		if (DO_DEBUG(debug_bit))			\
			VSL(SLT_Debug, (id), __VA_ARGS__);	\
	} while (0)

/* cache_seg_deliver.c */
void VSD_Init(void);

/* cache_vrt_re.c */
int VPI_re_lmatch(VRT_CTX, const char *s, int slen, void *re);

/* cache_vcl_vrt.c */
VCL_BACKEND VPI_LookupDirector(VRT_CTX, VCL_STRING name);

/* cache_vsync.c */
void SYNC_Init(void);
