/*-
 * Copyright (c) 2016-2017 Varnish Software AS
 * All rights reserved.
 *
 * Author: Reza Naghibi <reza@varnish-software.com>
 *
 */

#ifndef _VES_JSON_H_INCLUDED_
#define _VES_JSON_H_INCLUDED_

#include <errno.h>
#include <locale.h>
#include <math.h>
#include <float.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <wchar.h>

#ifndef ALLOC_OBJ
#include "miniobj.h"
#endif

#ifndef VDEF_H_INCLUDED
#include "vdef.h"
#endif

#include "vas.h"
#include "vsb.h"
#include "vtree.h"

#define VES_JSON_TOKEN_INIT_LEN		1024
#define VES_JSON_MAXSIZE		(1<<30)
#define VES_JSON_MAXDEPTH		64

#ifdef VES_STANDALONE
#define VES_JSON_PRINT_TFMT		"%s: '%.*s':%u"
#define VES_JSON_PRINT_TVAL(obj, tok)	ves_json_token_type((tok)->type), \
					(int)(tok)->raw_len, (tok)->buf, \
					(tok)->raw_len

#define ves_json_printf(args...)	_ves_json_printf(args)
#else
#define ves_json_printf(args...)
#endif

enum ves_json_object_state {
	VES_JSON_OBJECT_INIT = 0,
	VES_JSON_OBJECT_INDEXING,
	VES_JSON_OBJECT_NEEDMORE,
	VES_JSON_OBJECT_DONE,
	VES_JSON_OBJECT_ERROR,
	VES_JSON_OBJECT_BADJSON,
	VES_JSON_OBJECT_OOM
};

enum ves_json_token_type {
	VES_JSON_TOKEN_UNDEF = 0,
	VES_JSON_TOKEN_OBJECT,
	VES_JSON_TOKEN_ARRAY,
	VES_JSON_TOKEN_LABEL,
	VES_JSON_TOKEN_STRING,
	VES_JSON_TOKEN_NUMBER,
	VES_JSON_TOKEN_TRUE,
	VES_JSON_TOKEN_FALSE,
	VES_JSON_TOKEN_NULL
};

/* Note this is being used for object/array finalizing */
enum ves_json_layout {
	VES_JSON_LAYOUT_UNDEF = 0,
	VES_JSON_LAYOUT_INLINE,
	VES_JSON_LAYOUT_PTR
};

struct ves_json_token {
	unsigned			magic;
#define VES_JSON_TOKEN_MAGIC		0x223BE06C

	uint8_t				type;
	uint8_t				layout;

	uint32_t			raw_len;
	uint32_t			enc_len;

	uint32_t			size;
	uint32_t			parent;
	uint32_t			skip;

	uint32_t			left;
	uint32_t			right;

	char				buf[0];
};

struct ves_json_rbe {
	unsigned			magic;
#define VES_JSON_RBE_MAGIC		0x9B8AFE46

	VRB_ENTRY(ves_json_rbe)		entry;

	struct ves_json_object		*object;

	uint32_t			tindex;
};

VRB_HEAD(ves_json_rbtree, ves_json_rbe);

struct ves_json_parser {
	unsigned			magic;
#define VES_JSON_PARSER_MAGIC		0x404D17C9

	char				*buf;
	size_t				buf_len;

	size_t				offset;
	size_t				pos;

	size_t				context;
	size_t				context_len;
};

struct ves_json_index {
	unsigned			magic;
#define VES_JSON_INDEX_MAGIC		0x9AF237BE

	uint8_t				state;

	uint32_t			length;

	uint8_t				tokens[0];
};

struct ves_json_object {
	unsigned			magic;
#define VES_JSON_OBJECT_MAGIC		0xC3EA6202

	struct ves_json_parser		parser;

	size_t				alloc_len;

	const struct ves_json_index	*search;
	struct ves_json_index		*index;

#ifdef VARNISH_PLUS
	struct objcore			*oc;
	uintptr_t			oc_oaref;
	uintptr_t			json_oaref;
#endif
};

extern int ves_json_debug;

typedef void ves_json_output_f(void *data, const void *buf, size_t len);

/* Iterators */
#define	VES_JSON_FOREACH_FROM(obj, start, key, value)			\
	for ((key) = ves_json_itkv_init((obj), (start)),		\
	    (value) = ves_json_it_value((obj), (key));			\
	    ves_json_itkv_valid((obj), (start), (key), (value));	\
	    (key) = ves_json_skip_token((obj), (key)),			\
	    (value) = ves_json_it_value((obj), (key)))

#define	VES_JSON_ITER_FROM(obj, start, item)				\
	for ((item) = ves_json_it_init((obj), (start));			\
	    ves_json_it_valid((obj), (start), (item));			\
	    (item) = ves_json_skip_token((obj), (item)))

#define	VES_JSON_FOREACH(obj, key, value)				\
	VES_JSON_FOREACH_FROM((obj), NULL, (key), (value))

#define	VES_JSON_ITER(obj, item)					\
	VES_JSON_ITER_FROM((obj), NULL, (item))

/* Create JSON object */
struct ves_json_object *ves_json_object_alloc(void);
void ves_json_object_free(struct ves_json_object *object);
void ves_json_object_init(struct ves_json_object *object);

/* Create JSON index */
void ves_json_parse_chunk(struct ves_json_object *object, const char *ibuf,
    size_t buf_len);
void ves_json_finish(struct ves_json_object *object);

/* Import/export an index */
uint32_t ves_json_index_length(const struct ves_json_object *object);
const void *ves_json_index_get(const struct ves_json_object *object);
int ves_json_index_set(struct ves_json_object *object, const void *index,
    uint32_t len);

/* Searching and navigation */
const struct ves_json_token *ves_json_search(
    const struct ves_json_object *object, const struct ves_json_token *current,
    const char *search, size_t search_len, int depth);
const struct ves_json_token *ves_json_parent_token(
    const struct ves_json_object *object, const struct ves_json_token *token);
const struct ves_json_token *ves_json_skip_token(
    const struct ves_json_object *object, const struct ves_json_token *token);
const struct ves_json_token *ves_json_next_token(
    const struct ves_json_object *object, const struct ves_json_token *token);
const struct ves_json_token *ves_json_array_index(
    const struct ves_json_object *object, const struct ves_json_token *token,
    size_t i);
const struct ves_json_token *ves_json_head_token(
    const struct ves_json_object *object, const struct ves_json_token *token);
const struct ves_json_token *ves_json_search_token(
    const struct ves_json_object *object, const struct ves_json_token *token,
    const char *label, size_t len);
const struct ves_json_token *ves_json_root_token(
    const struct ves_json_object *object);
int ves_json_isroot_token(const struct ves_json_object *object,
    const struct ves_json_token *token);
const struct ves_json_token *ves_json_get_token(
    const struct ves_json_object *object, uint32_t tindex);
struct ves_json_token *ves_json_get_itoken(const struct ves_json_object *object,
    uint32_t tindex);
uint32_t ves_json_get_tindex(const struct ves_json_object *object,
    const struct ves_json_token *token);
struct ves_json_token *ves_json_context_token(
    const struct ves_json_object *object);
const struct ves_json_token * ves_json_it_init(
    const struct ves_json_object *object, const struct ves_json_token *start);
const struct ves_json_token *ves_json_itkv_init(
    const struct ves_json_object *object, const struct ves_json_token *start);
const struct ves_json_token *ves_json_it_value(
    const struct ves_json_object *object, const struct ves_json_token *key);
int ves_json_it_valid(const struct ves_json_object *object,
    const struct ves_json_token *start, const struct ves_json_token *item);
int ves_json_itkv_valid(const struct ves_json_object *object,
    const struct ves_json_token *start, const struct ves_json_token *key,
    const struct ves_json_token *value);

/* Utility */
enum ves_json_object_state ves_json_state(const struct ves_json_object *object);
enum ves_json_token_type ves_json_type(const struct ves_json_token *token);
size_t ves_json_token_len(const struct ves_json_token *token);
const char *ves_json_string_buf(const struct ves_json_token *token,
    size_t *string_len, int decode);
uint32_t ves_json_token_size(const struct ves_json_token *token, int align);
void ves_json_print_token(const struct ves_json_object *object,
    const struct ves_json_token *token,
    ves_json_output_f *callback, void *data, int decode, int quoted);
int ves_json_token_cmp(const char *s1, size_t len1, const char *s2,
    size_t len2);
const char *ves_json_token_type(enum ves_json_token_type type);
const char *ves_json_object_state(enum ves_json_object_state state);
const struct ves_json_index * ves_json_error_index(
    enum ves_json_object_state state);
const struct ves_json_token * ves_json_static_token(
    enum ves_json_token_type type);
enum ves_json_token_type ves_json_compare_type(
    const struct ves_json_token *token_lhs,
    const struct ves_json_token *token_rhs);
int ves_json_is_null(const struct ves_json_token *token);
int ves_json_is_truthy(const struct ves_json_token *token);
double ves_json_get_double(const struct ves_json_token *token, int *error);
double ves_json_str_to_double(const char *buf, size_t len, int *error);
size_t ves_json_to_string(char *out, size_t out_len, const char *in,
    size_t in_len);
void ves_json_to_vsb(struct vsb *out, const char *in, size_t in_len);
size_t ves_json_print(const struct ves_json_object *object,
    const struct ves_json_token *token, char *buf, size_t buf_len);
void ves_json_enable_codepoints(void);

#ifdef VES_STANDALONE
void _ves_json_printf(const char *format, ...)
    __attribute__((format(printf, 1, 2)));
void ves_json_object_printd(const struct ves_json_object *object);
void ves_json_object_print(const struct ves_json_object *object);
struct ves_json_object *ves_load_json_file(char *file);
struct ves_json_object *ves_load_json_mfile(const char *path, size_t path_len);
#endif

#endif	/* _VES_JSON_H_INCLUDED_ */
